#include <math.h> 
#include <stdio.h> 
#include "../ansi_lib/A_comp.h"
#include "../ansi_lib/io_mat.h"
#include "../ansi_lib/allocate.h"
#include "../ansi_lib/assort.h"
#include "../ansi_lib/fft.h"

#define PI        3.14159265358979323846

double hrl(int m, struct sino_info *sino);

int VERBOSE;

void main(int argc, char **argv)
{
  MATFile *fout,*fin;
  float **Y, **dosage;
  double *ans, *data, *filter, alpha, factor, temp, norm=0.0;
  int   nr,ntheta;
  float **image;
  int   N,M;
  int   i,j,t,ii,n,m, first = 1;
  struct column sp_col;
  struct sino_info sino;
  struct geom_info geom;
  float **profiles;
  int Nsup;
  double pmass, imass,mx;


  if (argc != 5){
    printf("Incorrect no. of arguments\n");
    printf("Usage: \n");
    printf("trans input.mat output.mat alpha VERBOSE\n");
    exit(-1);
  }

  sscanf(argv[3],"%lf",&alpha);
  sscanf(argv[4],"%d",&VERBOSE);


  fin = matOpen(argv[1], "r");
  if (fin == NULL) {
    fprintf(stderr,"Cannot read MatLab file %s \n",argv[1]);
    exit(-1);
  }

  dosage = read_mat_float_image(&ntheta,&nr,"dosage",fin);
  if (dosage == NULL) {
    fprintf(stderr,"Cannot read dosage \n");
  }
  Y = read_mat_float_image(&ntheta,&nr,"counts",fin);
  if (Y == NULL) {
    fprintf(stderr,"Cannot read counts matrix \n");
  }
  read_mat_sino_info(fin, &sino);
  matClose(fin);



    geom.dia = sino.n_t*sino.delta_t;
    geom.n_y = sino.n_t;
    geom.n_x = sino.n_t;
    geom.delta = sino.delta_t;
    geom.x_0 = (-geom.dia+geom.delta)/2.0;
    geom.y_0 = (-geom.dia+geom.delta)/2.0;

    pix_prof_comp(&sino, &geom, &profiles);
    sp_col.val = (float *) get_spc(sino.n_t*sino.n_theta, sizeof(float));
    sp_col.row = (int *) get_spc(sino.n_t*sino.n_theta, sizeof(int));

    N = geom.n_y;
    M = geom.n_x;




  for(i=0;i<ntheta;i++)
	 for(j=0;j<nr;j++){
	   temp = Y[i][j];
	   if (temp <= 0.0)
		 temp = 1.0;
	   Y[i][j] = log( dosage[0][0]/temp );
	   if (Y[i][j] < 0.0)
		 Y[i][j] = 0.0;
     }




  /********* Allocate space for image ************/
  image = (float **)get_img(M, N, sizeof(float));
  for(i=0; i<N; i++)
     for(j=0; j<M; j++) {
        image[i][j] = 0.0;
     }
  /***********************************************/

  Nsup = N*M;

  pmass = mean(Y,ntheta,nr)*nr*sino.delta_t/(N*M*geom.delta*geom.delta);




  n = (int) ceil(log(nr)/log(2.0));
  n = pow(2,n);


    /****** Allocate space for filter and data *******/

    ans = (double *)get_spc(4*n, sizeof(double));
    ans = ans -1;

    data = (double *)get_spc(2*n, sizeof(double));
    data = data - 1;

    filter = (double *)get_spc(2*n, sizeof(double));
    filter = filter - 1;




  /* Generalized Hamming filter */
  if(VERBOSE)
  {
    printf("alpha is %f\n",alpha);
    printf("Filtering data ... \n");
  }
  factor = (1- alpha)/2.0;
  for(i=0;i<=n;i++)
    filter[i+1] = alpha*hrl(i,&sino) + factor*( hrl(i-1,&sino) + hrl(i+1,&sino) );
  for(i=n+2;i<=2*n;i++)
    filter[i] = filter[2*n-i+2];

  /* Pad data with zeros */
  for(i=nr+1;i<=2*n;i++)
    data[i] = 0.0;


  /* Filter data for each projection angle */
  for(i=0;i<ntheta;i++) {

    for(j=0;j<nr;j++)
      data[j+1] = Y[i][j];

      convlv(data,2*n,filter,ans);

      for(j=0;j<nr;j++)
        Y[i][j] = ans[j+1];
  }



  if(VERBOSE)
    printf("Backprojecting ... \n");

  /* backprojection */
  for(i=0; i<N; i++)
  for(j=0; j<M; j++) {
    A_comp(i,j,&sino, &geom, profiles,&sp_col);
	norm = 0.0;
    for(t=0; t<sp_col.n_row; t++) {
      ii = sp_col.row[t];
      image[i][j] += sp_col.val[t]*Y[0][ii];
      norm += sp_col.val[t];
    }
    if (image[i][j] <= 0.0 )
      image[i][j] = 0.0;
    else
      image[i][j] = image[i][j]*PI/(norm);
  }


  imass = 0.0;
  for (i=0;i<N*M;i++)
	imass += image[0][i];
  imass /= (N*M);

  if(VERBOSE)
    printf("Mean value correction ... \n");

  for (i=0;i<N*M;i++)
	image[0][i] *= (pmass/imass);

  mx = 0.0;
  for (i=0;i<N*M;i++)
	mx += image[0][i];

  mx /= Nsup;


  if(VERBOSE)
  {
  printf("The mean of the CBP image is %f\n",mx);
  printf("The min,max of the CBP image are %.3f,%.3f\n",
          min(image,N,M),max(image,N,M));
  }



  fout = matOpen(argv[2], "w");
  if (fout == NULL) {
    fprintf(stderr,"Cannot write to MatLab file %s\n",argv[2]);
    exit(-1);
  }

  write_mat_float_image(image,N,M,"img",fout);
  close(fout);

    free((void *)sp_col.val);
    free((void *)sp_col.row);
    free((void *)(ans+1));
    free((void *)(data+1));
    free((void *)(filter+1));
    free_img((void **)profiles);
    free_img((void **)image);
    free_img((void **)Y);

}



double hrl(int m, struct sino_info *sino)
{
    float temp;

    if (m == 0)
        return(1/(4.0*sino->delta_t));
    else
    {
        temp = PI*m/2;
        temp = sin(temp)/temp;
        temp =  - temp*temp;
        return( temp/(4.0*sino->delta_t) );
    }
}


