#include <stdio.h>
#include <math.h>
#include "../ansi_lib/io_mat.h"
#include "../ansi_lib/A_comp.h"
#include "../ansi_lib/allocate.h"
#include "../ansi_lib/assort.h"
#include "../ansi_lib/multigrid.h"
#include "../ansi_lib/mrf_emis_otf.h"


int VERBOSE;


struct column sp_col;

double scatter;


main( int argc, char **argv )
{

    MATFile *fout,*fin;

	#ifdef MATLAB5
	  mxArray *pm;
    #else
	  Matrix *pm;
    #endif

	int N,M,Ang,T,i,j;
    int no_iter_param,no_iter_rec;
	char matrix[20];
	struct sino_info sino;
	struct geom_info geom;
	float **temp_image, p;
    float **X;
    float **Y;
    float **sigma;
	int **mask;

    if (argc != 9)
    {
        printf("Incorrect no. of arguments\n");
        printf("Usage:\n");
        printf("emis input.mat initial.mat support.mat p no_iter_param no_iter_rec output.mat VERBOSE\n");
        exit(-1);
    }
    sscanf(argv[4],"%f",&p);
    sscanf(argv[5],"%d",&no_iter_param);
    sscanf(argv[6],"%d",&no_iter_rec);
    sscanf(argv[8],"%d",&VERBOSE);
	if(VERBOSE)
      printf("p %f, no_iter_param %d no_iter_rec %d\n",p,no_iter_param,no_iter_rec);


	srandom2(1);

    fin = matOpen(argv[1], "r");
    if (fin == NULL)
    {
        fprintf(stderr,"Cannot read MatLab file %s\n",argv[1]);
        exit(-1);
    }

    Y = read_mat_float_image(&Ang,&T,"counts",fin);
    if (Y == NULL)
    {
        fprintf(stderr,"Cannot read sino matrix \n");
        exit(-1);
    }
	read_mat_sino_info(fin, &sino);

    #ifdef MATLAB5
    pm = matGetArray(fin,"scatter");
    if(pm != NULL) {
      scatter = mxGetScalar(pm);
	}
    #else
    pm = matGetMatrix(fin,"scatter");
    if(pm != NULL) {
      scatter = mxGetScalar(pm);
    }
    #endif

    if(scatter == 0.0)
      scatter = 0.001;

	matClose(fin);

    fin = matOpen(argv[2], "r");
    if (fin == NULL)
    {
        fprintf(stderr,"Cannot read MatLab file %s\n",argv[2]);
        exit(-1);
    }

    X = read_mat_float_image(&N,&M,"img",fin);
    if (X == NULL)
    {
        fprintf(stderr,"Cannot read img matrix \n");
        exit(-1);
    }
    matClose(fin);


    fin = matOpen(argv[3], "r");
    if (fin == NULL)
    {
        fprintf(stderr,"Cannot read MatLab file %s\n",argv[3]);
        exit(-1);
    }

    temp_image = read_mat_float_image(&N,&M,"mask",fin);
    if (temp_image == NULL)
    {
        fprintf(stderr,"Cannot read mask matrix \n");
        exit(-1);
    }
    matClose(fin);



	geom.dia = sino.n_t*sino.delta_t;
	geom.n_y = sino.n_t;
	geom.n_x = sino.n_t;
	geom.delta = sino.delta_t;
 	geom.x_0 = (-geom.dia+geom.delta)/2.0;
 	geom.y_0 = (-geom.dia+geom.delta)/2.0;



	N=geom.n_y;
	M=geom.n_x;

	/* Allocate memory and initialize projection matrices*/

    sp_col.val = (float *) get_spc(sino.n_t*sino.n_theta, sizeof(float));
    sp_col.row = (int *) get_spc(sino.n_t*sino.n_theta, sizeof(int));
	sp_col.multiplier = 1.0;

    sigma = (float **)get_img(1,1,sizeof(float));
    mask = (int **)get_img(M,N,sizeof(int));
    for(i=0;i<N*M;i++)
        mask[0][i] = (int)temp_image[0][i];


    free_img((void **)temp_image);



	sigma[0][0] = mrf_emis_otf(X,mask,Y,&sino,&geom,no_iter_param,no_iter_rec,p);

    if(VERBOSE)
		printf("Writing output image ...\n");


    fout = matOpen(argv[7], "w");
    if (fout == NULL)
    {
        fprintf(stderr,"Cannot write MatLab file %s\n",argv[7]);
        exit(-1);
    }
	sprintf(matrix,"img");
   	write_mat_float_image(X,N,M,matrix,fout);
	sprintf(matrix,"sigma");
   	write_mat_float_image(sigma,1,1,matrix,fout);
    matClose(fout);


    if(VERBOSE)
		printf("Freeing memory ...\n");

    free_img((void **)X);
    free_img((void **)mask);
    free_img((void **)sigma);
    free_img((void **)Y);
    free((void *)sp_col.val);
    free((void *)sp_col.row);


		
	
}
