#include <stdio.h>
#include <math.h>
#include "../ansi_lib/io_mat.h"
#include "../ansi_lib/A_comp.h"
#include "../ansi_lib/allocate.h"
#include "../ansi_lib/assort.h"
#include "../ansi_lib/multigrid.h"
#include "../ansi_lib/mr_mrf_emis_otf.h"
#include "../ansi_lib/mrf_emis_otf.h"


int VERBOSE;

#define MAX_k 20

void free_mem(void);
int level(int N);


float ***X, ***RX, **SX[MAX_k];
int ***mask;
float **Y;
int no_levels;

struct column sp_col;

double scatter;


main( int argc, char **argv )
{

    MATFile *fout,*fin;

    #ifdef MATLAB5
      mxArray *pm;
    #else
      Matrix *pm;
    #endif

	int N,M,Ang,T,Nk,Mk,k,i,j,No_pix;
    int iter,no_iter;
	char matrix[20],c;
	struct sino_info sino;
	struct geom_info geom[MAX_k];
	float **temp_image;
    float p;

	double m;

    if (argc != 7)
    {
        printf("Incorrect no. of arguments\n");
        printf("Usage:\n");
        printf("emis input.mat support.mat p no_iter output.mat VERBOSE\n");
        exit(-1);
    }

	sscanf(argv[3],"%f",&p);
	sscanf(argv[4],"%d",&iter);
	sscanf(argv[6],"%d",&VERBOSE);


	srandom2(1);

    fin = matOpen(argv[1], "r");
    if (fin == NULL)
    {
        fprintf(stderr,"Cannot read MatLab file %s\n",argv[1]);
        exit(-1);
    }

    Y = read_mat_float_image(&Ang,&T,"counts",fin);
    if (Y == NULL)
    {
        fprintf(stderr,"Cannot read sino matrix \n");
        exit(-1);
    }
	read_mat_sino_info(fin, &sino);
    #ifdef MATLAB5
    pm = matGetArray(fin,"scatter");
    if(pm != NULL) {
      scatter = mxGetScalar(pm);
    }
    #else
    pm = matGetMatrix(fin,"scatter");
    if(pm != NULL) {
      scatter = mxGetScalar(pm);
    }
    #endif
    if(scatter == 0.0)
      scatter = 0.001;
	matClose(fin);

    fin = matOpen(argv[2], "r");
    if (fin == NULL)
    {
        fprintf(stderr,"Cannot read MatLab file %s\n",argv[2]);
        exit(-1);
    }

    temp_image = read_mat_float_image(&N,&M,"mask",fin);
    if (temp_image == NULL)
    {
        fprintf(stderr,"Cannot read mask matrix \n");
        exit(-1);
    }
    matClose(fin);



	geom[0].dia = sino.n_t*sino.delta_t;
	geom[0].n_y = sino.n_t;
	geom[0].n_x = sino.n_t;
	geom[0].delta = sino.delta_t;
 	geom[0].x_0 = (-geom[0].dia+geom[0].delta)/2.0;
 	geom[0].y_0 = (-geom[0].dia+geom[0].delta)/2.0;



	N=geom[0].n_y;
	M=geom[0].n_x;

	m = mean(Y,Ang,T)*T*sino.delta_t/(N*M*geom[0].delta*geom[0].delta);
	printf("mean obtained from projections %f\n",m);

	if( N < M )
		no_levels = level(N);
	else
		no_levels = level(M);

    No_pix = N*M/pow(4.0,no_levels-1);
    if( No_pix < 64 && no_levels > 3)
    {
        i = log(64.0/(double) No_pix)/log(4.0);
        no_levels = no_levels-i;
    }


    if(VERBOSE)
		printf("No. of levels %d\n",no_levels);

	/* Allocate memory and initialize projection matrices*/

    sp_col.val = (float *) get_spc(sino.n_t*sino.n_theta, sizeof(float));
    sp_col.row = (int *) get_spc(sino.n_t*sino.n_theta, sizeof(int));
	sp_col.multiplier = 1.0;

	X = (float ***) get_spc(no_levels,sizeof(float **)); 
	RX = (float ***) get_spc(no_levels,sizeof(float **)); 
	mask = (int ***) get_spc(no_levels,sizeof(int **)); 

	Nk = N;
	Mk = M;
	for(k=0;k<no_levels;k++)
	{

		X[k] = (float **)get_img(Mk,Nk,sizeof(float));
		RX[k] = (float **)get_img(Mk,Nk,sizeof(float));
		mask[k] = (int **)get_img(Mk,Nk,sizeof(int));

		if(k==0)
          for(i=0;i<N*M;i++)
            mask[0][0][i] = (int)temp_image[0][i];
		else
		{
		    geom[k].dia = sino.n_t*sino.delta_t;
    		geom[k].n_y = Nk;
    		geom[k].n_x = Mk;
    		geom[k].delta = geom[k-1].delta*2.0;
		    geom[k].x_0 = (-geom[k].dia+geom[k].delta)/2.0;
		    geom[k].y_0 = (-geom[k].dia+geom[k].delta)/2.0;
			decimate_mask(mask[k],Nk,Mk,mask[k-1]);
		}	


		if (k == no_levels-1)
			for(i=0;i<Nk*Mk;i++)
				X[k][0][i] = m;

		Nk = Nk/2;
		Mk = Mk/2;
	}

    free_img((void **)temp_image);



    for(k=no_levels-1;k>=0;k--)
	{
		no_iter = (int) ceil((double)iter*pow(2.0,(double)k/3.0));

		Nk = geom[k].n_y;
		Mk = geom[k].n_x;

        if( k != no_levels-1)
        {
        	interp(X[k+1],Nk/2,Mk/2,RX[k],1);
			for(i=0;i<Mk*Nk;i++)
				X[k][0][i] = RX[k][0][i];
		}

		printf("level %d no_iter %d\n",k,no_iter);

		if(k==no_levels-1 )
			mrf_emis_otf(X[k],mask[k],Y,&sino,&geom[k],5,no_iter,p);
		else
			SX[k] = mr_mrf_emis_otf(X[k],RX[k],mask[k],Y,&sino,&geom[k],3,no_iter,p);

	}




    fout = matOpen(argv[5], "w");
    if (fout == NULL)
    {
        fprintf(stderr,"Cannot write MatLab file %s\n",argv[5]);
        exit(-1);
    }
	for(k=0;k<no_levels;k++)
	{
		Nk = geom[k].n_y;
		Mk = geom[k].n_x;
		sprintf(matrix,"X%d",k);
    	write_mat_float_image(X[k],Nk,Mk,matrix,fout);
		sprintf(matrix,"RX%d",k);
    	write_mat_float_image(RX[k],Nk,Mk,matrix,fout);
		if(k != no_levels-1 )
		{
			sprintf(matrix,"SX%d",k);
    		write_mat_float_image(SX[k],Nk,Mk,matrix,fout);
		}
	}
    close(fout);

    if(VERBOSE)
		printf("Freeing memory ...\n");

	free_mem();

		
	
}

void free_mem(void)
{
    int k,i;


    for(k=0;k<no_levels;k++)
    {
        free_img((void **)X[k]);
        free_img((void **)RX[k]);
        free_img((void **)mask[k]);
		if( k != no_levels-1)
        	free_img((void **)SX[k]);
    }
    free((void *)X);
    free((void *)RX);
    free((void *)mask);
    free_img((void **)Y);
    free((void *)sp_col.val);
    free((void *)sp_col.row);

}

int level(int N)
{
  int k=0,div=1;

  while( div*(N/div) == N )
  {
    div *= 2;
    k++;
  }
  return(k);
}

