#include<stdio.h>
#include<math.h>
#include <stdlib.h>
#include "A_comp.h"
#include "allocate.h"

#define LEN 512

extern int VERBOSE;



void pix_prof_comp(
 struct sino_info *sino,
 struct geom_info *geom,
 float ***pix_prof)
{

 int	i,j;
 double pi,ang,d1,d2,t,t_1,t_2,t_3,t_4,maxval,scale,rc;

 *pix_prof = (float **)get_img(LEN,sino->n_theta,sizeof(float));

 pi = 4.0 * atan(1.0);

/* Fill profiles of pixels from all angles
 Each profile assumed 2 pixels wide
 This is used to speed calculation (too many to store) of the
        entries of the projection matrix.
 Values are scaled by "scale"
*/
/* WATCH THIS; ONLY FOR SQUARE PIXELS NOW 	*/

	scale = geom->delta;
	/*
	printf("delta_x = %f, delta_theta = %f\n",scale,sino->delta_theta);*/

    rc = sin((double)(pi/4.0));
    for(i = 0; i < sino->n_theta; i++) {
	 ang = sino->theta_0+i*sino->delta_theta;

	 while(ang >= pi/2.0)
		ang -= pi/2.0;
         
         while(ang < 0.0)
                ang += pi/2.0;

         if(ang <= pi/4.0)
                maxval = scale/cos((double) ang);
         else
                maxval = scale/cos((double) (pi/2.0-ang));

         d1 = rc * cos((double) (pi/4.0-ang));
         d2 = rc*fabs((double) (sin((double)(pi/4.0-ang))));
         t_1 = 1.0-d1;
         t_2 = 1.0-d2;
         t_3 = 1.0+d2;
         t_4 = 1.0+d1;
         for(j = 0; j < LEN; j++) {
                t = 2.0*j / (double)LEN;
                if(t<=t_1 || t>=t_4)
                        (*pix_prof)[i][j] = 0.0;
                else if(t>t_3)
                        (*pix_prof)[i][j] = maxval*(t_4-t)/(t_4-t_3);
                else if(t>=t_2)
                        (*pix_prof)[i][j] = maxval;
                else
                        (*pix_prof)[i][j] = maxval*(t-t_1)/(t_2-t_1);
         }
    }
	/*
	printf("Finished profiles computation\n");*/


} /* Close prof_comp()	*/

void A_comp(
 int im_row, int im_col,
 struct sino_info *sino,
 struct geom_info *geom,
 float **pix_prof,
 struct column *sp_col)

{
 int	ind,ind_min,ind_max,pr,
	prof_ind,pind,i,j,proj_count;

 double	Aval,t_min,t_max,ang,x,y;

/* WATCH THIS; ONLY FOR SQUARE PIXELS NOW 	*/

 
  y = geom->y_0 + im_row*geom->delta;
  x = geom->x_0 + im_col*geom->delta;

   proj_count = 0;
   for(pr=0; pr < sino->n_theta; pr++)
   {
    ang = sino->theta_0 + pr*(sino->delta_theta);
    pind = pr*sino->n_t;

   /* Need profile to contain 2 pixel widths       */
    t_min = y*cos((double) ang) - x*sin((double) ang) - geom->delta;
    t_max = t_min + 2.0*geom->delta;
    if(t_max < sino->t_0) /* This also prevents over-reach (with rounding of
		negative numbers)	*/
	  continue;

    ind_min = (t_min/geom->dia+0.5)*sino->n_t + 0.5;
    ind_max = (t_max/geom->dia+0.5)*sino->n_t - 0.5;

   /* Fix this 4/91 to prevent over-reach at ends  */
    ind_min = (ind_min<0) ? 0:ind_min;
    ind_max = (ind_max>=sino->n_t) ? sino->n_t-1:ind_max;
    for(i = ind_min; i <=ind_max; i++) {

 	  ind = pind+i;
 	  prof_ind = LEN*(sino->t_0+i*sino->delta_t-t_min)/(2.0*geom->delta);

	  if(prof_ind >= LEN || prof_ind <0){
	  /*
		printf("Error in index computation for profiles %d\n",prof_ind);
		*/
		if(prof_ind == LEN)
			prof_ind = LEN-1;
		else if (prof_ind == -1)
			prof_ind = 0;
		else {
			fprintf(stderr,"\nExiting Program: input parameters inconsistant\n");
			exit(-1);
		}
	  }

	  if((Aval=pix_prof[pr][prof_ind]) > 0.0) {
 	    sp_col->val[proj_count] = Aval;
 	    sp_col->row[proj_count] = ind;/* Index for ray    */
	    proj_count++;
      }
	}
   } 
   sp_col->n_row = proj_count;
   /*
   printf("proj_count = %d\n",proj_count);
   */


}/* Close A_comp()	*/

#undef LEN 

#include <mat.h>


#ifdef MATLAB5

#include <matrix.h>

void read_mat_sino_info(
  MATFile *fin,       /* pointer to file */
  struct sino_info *sino
  )
{
 mxArray *pm;

 pm = matGetArray(fin,"n_theta");
 if(pm != NULL){
  sino->n_theta = (int) mxGetScalar(pm);
  }
 pm = matGetArray(fin,"n_t");
 if(pm != NULL){
  sino->n_t = (int) mxGetScalar(pm);
  }
 pm = matGetArray(fin,"delta_theta");
 if(pm != NULL){
  sino->delta_theta = mxGetScalar(pm);
  }
 pm = matGetArray(fin,"delta_t");
 if(pm != NULL){
  sino->delta_t = mxGetScalar(pm);
  }
 pm = matGetArray(fin,"theta_0");
 if(pm != NULL){
  sino->theta_0 = mxGetScalar(pm);
  }
 pm = matGetArray(fin,"t_0");
 if(pm != NULL){
  sino->t_0 = mxGetScalar(pm);
  }

 sino->dec_a = 1;
 sino->dec_t = 1;

 if(VERBOSE)
 {
 printf("delta_t = %f\n",sino->delta_t);
 printf("delta_theta = %f\n",sino->delta_theta);
 printf("t_0 = %f\n",sino->t_0);
 printf("theta_0 = %f\n",sino->theta_0);
 printf("n_t = %d\n",sino->n_t);
 printf("n_theta = %d\n",sino->n_theta);
 }

}

#else

#include <mat.h>

void read_mat_sino_info(
  MATFile *fin,       /* pointer to file */
  struct sino_info *sino
  )
{
 Matrix *pm;

 pm = matGetMatrix(fin,"n_theta");
 if(pm != NULL){
  sino->n_theta = (int) mxGetScalar(pm);
  }
 pm = matGetMatrix(fin,"n_t");
 if(pm != NULL){
  sino->n_t = (int) mxGetScalar(pm);
  }
 pm = matGetMatrix(fin,"delta_theta");
 if(pm != NULL){
  sino->delta_theta = mxGetScalar(pm);
  }
 pm = matGetMatrix(fin,"delta_t");
 if(pm != NULL){
  sino->delta_t = mxGetScalar(pm);
  }
 pm = matGetMatrix(fin,"theta_0");
 if(pm != NULL){
  sino->theta_0 = mxGetScalar(pm);
  }
 pm = matGetMatrix(fin,"t_0");
 if(pm != NULL){
  sino->t_0 = mxGetScalar(pm);
  }

 sino->dec_a = 1;
 sino->dec_t = 1;

 if(VERBOSE)
 {
 printf("delta_t = %f\n",sino->delta_t);
 printf("delta_theta = %f\n",sino->delta_theta);
 printf("t_0 = %f\n",sino->t_0);
 printf("theta_0 = %f\n",sino->theta_0);
 printf("n_t = %d\n",sino->n_t);
 printf("n_theta = %d\n",sino->n_theta);
 }

}

#endif
