#include <stdio.h>
#include <math.h>
#include "tdf/tdf.h"
#include "allocate.h"
#include "assort.h"
#include "solve.h"
#include "randlib.h"
#include "mr_mrf_emis.h"

#define TOL 0.00001
#define WN 0.14644660940673
#define WD 0.10355339059327
#define THRESHOLD 0.0


#define SIGN(A,B) ((A) > (B) ? 1.0 : -1.0)

static void f_pro(float **X, struct system_matrix *A);
static double f(double x);
static double df(double x);
static void relax(float **X, float **Y, struct system_matrix *A, int no_iter);
static double cal_SIGMA(float **X, struct system_matrix *A);
static void set_SG(float **RX,struct system_matrix *A);

static float updateX(void);
static void minmax(float *low, float *high);


static float *S, **SG;
static int param, Nsup;

static float SIGMA,CX,SXN[4],SXD[4],ML,NX[4],DX[4],P;
static double D1, D2;


float **mr_mrf_emis(float **X, float **RX, float **Y, struct system_matrix *A, int no_iter_param, int no_iter_rec, float p, float *start_SIGMA)
{


	int N,M,k,i,j;



	srandom2(1);
	P = p-1;
	SIGMA = *start_SIGMA;

	N = A->n_y;
	M = A->n_x;


	/* Allocate memory and initialize projection matrices*/

    S = (float *) get_spc(A->Mpro, sizeof(float));
	SG = (float **) get_img(M,N, sizeof(float));


   	Nsup = 0;
   	for( i=0;i<A->Npix;i++ )
       	if ( A->n_row[i] )
		{
			if(X[0][i] > THRESHOLD)
           		Nsup++;
		}
		else
		{
			X[0][i] = 0.0;
			RX[0][i] = 0.0;
		}

	set_SG(RX,A);



	f_pro(X,A);
	param = 1;
    relax(X, Y, A, no_iter_param);

	for( i=0;i<A->Npix;i++ )
		X[0][i] = RX[0][i];

	f_pro(X,A);
	param = 0;
    relax(X, Y, A, no_iter_rec);



   	for( i=0;i<A->Npix;i++ )
		SG[0][i] = SG[0][i]*(P+1)*SIGMA;

	*start_SIGMA = SIGMA;

	free((void *)S);

	return(SG);

}

static void set_SG(float **RX,struct system_matrix *A)
{
    int i,j,ai,aj,si,sj,N,M;
    float temp;

    N = A->n_y;
    M = A->n_x;


    for (i=0; i<N; i++)
        for(j=0; j<M; j++)
        {
                ai = (i+1)%N;
                aj = (j+1)%M;
                si = i-1;
                if (si < 0)
                    si = N-1;
                sj = j-1;
                if (sj < 0)
                    sj = M-1;

                temp = RX[i][j];
                SG[i][j] = WN*(pow(fabs(temp - RX[i][aj]),P+1) + 
				                pow(fabs(temp - RX[i][sj]),P+1) + 
								pow(fabs(temp - RX[ai][j]),P+1) + 
								pow(fabs(temp - RX[si][j]),P+1));

                SG[i][j] += WD*(pow(fabs(temp - RX[ai][aj]),P+1) + 
								 pow(fabs(temp - RX[ai][sj]),P+1) + 
								 pow(fabs(temp - RX[si][aj]),P+1) + 
								 pow(fabs(temp - RX[si][sj]),P+1));


/*
				if(SG[i][j] == 0.0)
					SG[i][j] = 1.0e-8;
					*/
				if(SG[i][j] < 1.0e-3)
					SG[i][j] = 1.0e-3;
        }
}

static double cal_SIGMA(float **X, struct system_matrix *A)
{
    int i,j,N,M;
    int aj,sj,ai;
    double sum = 0.0, sumd = 0.0;
	double temp;

	N = A->n_y;
	M = A->n_x;

	Nsup = 0;
    for(j=0;j<M;j++)
        for(i=0;i<N;i++)
			if(A->n_row[i*M+j])
            {
				if(X[i][j] > THRESHOLD)
					Nsup++;

                ai = (i+1)%N;
                aj = (j+1)%M;
                sj = j-1;
                if (sj < 0)
                    sj = M-1;
    
    
                temp = (SG[i][j] + SG[ai][j])/2.0;
                sum += pow(fabs(X[i][j] - X[ai][j]),P+1)/temp;
                temp = (SG[i][j] + SG[i][aj])/2.0;
                sum += pow(fabs(X[i][j] - X[i][aj]),P+1)/temp;
                temp = (SG[i][j] + SG[ai][sj])/2.0;
                sumd += pow(fabs(X[i][j] - X[ai][sj]),P+1)/temp;
                temp = (SG[i][j] + SG[ai][aj])/2.0;
                sumd += pow(fabs(X[i][j] - X[ai][aj]),P+1)/temp;
            }

    sum = sum*WN;
    sumd = sumd*WD;

    return((sum+sumd));
}




static void relax(float **X, float **Y, struct system_matrix *A, int no_iter)
{

	int iter;
	int i,j,N,M,pixel,pro,t;
	int ai,aj,si,sj;
	float diff;
	double l_like, temp;
	char c;

	N = A->n_y;
	M = A->n_x;




	if (SIGMA == 0.0)
	{
		SIGMA = cal_SIGMA(X, A)/Nsup;
		if (SIGMA == 0.0)
		{
			SIGMA = 1.0;
			printf("Warning - SIGMA arbitrarily initialized to 1.0\n");
		}
	} 

    printf("Starting sigma %f Nsup %d\n",pow(SIGMA,1.0/(P+1)),Nsup);

	for(iter=0;iter<no_iter;iter++)
    {

        for (i=0; i<N; i++)
            for(j=0; j<M; j++)
            {

                pixel = i*M+j;
                if(A->n_row[pixel])
                {


                    D1 = 0.0;
                    D2 = 0.0;
                    for(t=0;t<A->n_row[pixel];t++)
                    {
                        pro = A->row[pixel][t];
                        if (S[pro] != 0.0)
                        {
                            D1 += A->val[pixel][t]*(1.0 - Y[0][pro]/S[pro]);
                            D2 += Y[0][pro]*sq(A->val[pixel][t]/S[pro]);

                        }
                        else if(Y[0][pro] != 0.0)
                            printf("Projection values don't agree !!\n");


                    }




                    CX = X[i][j];
                    ML = CX - D1/D2;

                    ai = (i+1)%N;
                    aj = (j+1)%M;
                    si = i-1;
                    if (si < 0)
                        si = N-1;
                    sj = j-1;
                    if (sj < 0)
                        sj = M-1;

					SXN[0] = (SG[i][j] + SG[i][aj])/2.0;
					SXN[1] = (SG[i][j] + SG[i][sj])/2.0;
					SXN[2] = (SG[i][j] + SG[ai][j])/2.0;
					SXN[3] = (SG[i][j] + SG[si][j])/2.0;

					SXD[0] = (SG[i][j] + SG[ai][aj])/2.0;
					SXD[1] = (SG[i][j] + SG[ai][sj])/2.0;
					SXD[2] = (SG[i][j] + SG[si][aj])/2.0;
					SXD[3] = (SG[i][j] + SG[si][sj])/2.0;

                    NX[0] = X[i][aj];
                    NX[1] = X[i][sj];
                    NX[2] = X[ai][j];
                    NX[3] = X[si][j];

                    DX[0] = X[ai][aj];
                    DX[1] = X[ai][sj];
                    DX[2] = X[si][aj];
                    DX[3] = X[si][sj];


					X[i][j] = updateX();
					diff = X[i][j] - CX;
                	if(diff)
                    	for(t=0;t<A->n_row[pixel];t++)
                    	{
                        	pro = A->row[pixel][t];
                        	S[pro] += A->val[pixel][t]*diff;
                    	}

				}



            }

        if (param)
        {
			SIGMA = cal_SIGMA(X, A)/Nsup;
            printf("sigma %f Nsup %d\n",pow(SIGMA,1.0/(P+1)),Nsup);
        }
        else
        {
            l_like = 0.0;
            for(i=0;i<A->Mpro;i++)
                if ( S[i] > 0.0 )
                    l_like += -S[i] + Y[0][i]*log( S[i] );

            l_like -= cal_SIGMA(X,A)/((P+1)*SIGMA);
            printf("l_like %f\n",l_like);
        }

	}

}	


							
			




static void minmax(float *low, float *high)
{
    int i;

    *low = ML;
    *high = ML;

    for(i=0;i<4;i++)
    {

        if( *low > NX[i])
            *low = NX[i];
        else if(*high < NX[i])
            *high = NX[i];

        if( *low > DX[i])
            *low = DX[i];
        else if(*high < DX[i])
            *high = DX[i];

    }
}


static float updateX(void)
{
	float mm,new,std,var,beta,temp;
	float correction,dE,pert,threshold,low,high;
	int status;


    minmax(&low, &high);

    mm = solve(df, low, high, TOL, &status);
    if (status)
    {
        if (f(low) < f(high))
            mm = low;
        else
            mm = high;
    }

	if (param)
	{
		var = 1/D2;
		if (var < 0.0)
		{	
			printf("var. negative\n");
			exit(-1);
		}
		std = sqrt(var);
		if ( mm > -std/4.0 )
		{
			while((new=std*normal() + mm)<0.0);
			correction = (CX-mm)*(CX-mm)/(2.0*var);
			correction -= (new-mm)*(new-mm)/(2.0*var);
		}
		else
		{
			if (df(0.0) <= 0.0)
			{	
				printf("df negative %f,%f,%f,%f\n",df(0.0),mm,ML,std);
				exit(-1);
			}

			beta = 10.0/(df(0.0));
			new = beta*exprand();
			correction = (CX-new)/beta;
		}

		dE = f(new) - f(CX) + correction;

		if (!(dE < 0.0 || (random2() < exp(-dE))))
			new = CX;

	}
	else
		if( mm < 0.0)
			new = 0.0;
		else
			new = mm;

	return(new);
}
	
	

static double df(double x)
{
    float prior=0.0, dprior=0.0;
    int i;

    for(i=0;i<4;i++)
    {
        prior += pow(fabs(x - NX[i]),P)*SIGN(x,NX[i])/SXN[i];
        dprior += pow(fabs(x - DX[i]),P)*SIGN(x,DX[i])/SXD[i];
    }
    prior  = prior*WN;
    dprior  = dprior*WD;

    return( D1 + D2*(x - CX) + (prior+dprior)/SIGMA );
}

static double f(double x)
{
    float temp, prior=0.0, dprior=0.0;
    int i;

    temp = x-CX;

    for(i=0;i<4;i++)
    {
        prior += pow(fabs(x - NX[i]),P+1)/SXN[i];
        dprior += pow(fabs(x - DX[i]),P+1)/SXD[i];
    }
    prior  = prior*WN;
    dprior  = dprior*WD;

    return( D1*temp + D2*temp*temp/2.0 + (prior + dprior )/((P+1)*SIGMA) );
}


static void f_pro(float **X, struct system_matrix *A)
{
    int i,j;

    for(i=0;i<A->Mpro;i++)
        S[i] = 0.0;

    for(j=0;j<A->Npix;j++)
        for(i=0;i<A->n_row[j]; i++)
            S[A->row[j][i]] += A->val[j][i]*X[0][j];
}


#undef TOL
#undef WN
#undef WD
#undef SIGN

