#include <stdio.h>
#include <math.h>
#include "allocate.h"
#include "assort.h"
#include "solve.h"
#include "randlib.h"
#include "A_comp.h"

#define TOL 1.0e-9 
#define WN 0.14644660940673
#define WD 0.10355339059327

extern double scatter;
extern int VERBOSE;


#define SIGN(A,B) ((A) > (B) ? 1.0 : -1.0)

static void f_pro(float **X, struct sino_info *sino, struct geom_info *geom, float **profiles);
static void relax(float **X, int **mask, float **Y, struct sino_info *sino, struct geom_info *geom, float **profiles, int no_iter);
static double f(double x);
static double df(double x);
static double cal_SIGMA(float **X, int **mask, int flag_l_like);
static void set_SG(float **RX);

static float updateX(void);
static void minmax(float *low, float *high);


static float *S, **SG;
static int param, Nsup, N, M, Mpro;

static float SIGMA=0.0,CX,SXN[4],SXD[4],ML,NX[4],DX[4],P;
static double D1, D2;
extern struct column sp_col;



float **mr_mrf_emis_otf(float **X, float **RX, int **mask, float **Y, struct sino_info *sino, struct geom_info *geom, int no_iter_param, int no_iter_rec, float p)
{
	int k,i,j;
	float **profiles;



	srandom2(1);
	P = p-1;

	N = geom->n_y;
	M = geom->n_x;
	Mpro = sino->n_t*sino->n_theta;
	pix_prof_comp(sino, geom, &profiles);

	Nsup=0;
	for(i=0;i<N*M;i++)
	  if(mask[0][i])
		Nsup++;


	/* Allocate memory and initialize projection matrices*/

    S = (float *) get_spc(Mpro, sizeof(float));
	SG = (float **) get_img(M,N, sizeof(float));


	set_SG(RX);



	f_pro(X,sino,geom,profiles);
	param = 1;
    relax(X, mask, Y, sino, geom, profiles, no_iter_param);

	for( i=0;i<N*M;i++ )
		X[0][i] = RX[0][i];

	f_pro(X,sino,geom,profiles);
	param = 0;
    relax(X, mask, Y, sino, geom, profiles, no_iter_rec);

	free((void *)S);
	free_img((void **)profiles);

	return(SG);

}

static void set_SG(float **RX)
{
    int i,j,ai,aj,si,sj;
    float temp;



    for (i=0; i<N; i++)
        for(j=0; j<M; j++)
        {
                ai = (i+1)%N;
                aj = (j+1)%M;
                si = i-1;
                if (si < 0)
                    si = N-1;
                sj = j-1;
                if (sj < 0)
                    sj = M-1;

                temp = RX[i][j];
                SG[i][j] = WN*(pow(fabs(temp - RX[i][aj]),P+1) + 
				                pow(fabs(temp - RX[i][sj]),P+1) + 
								pow(fabs(temp - RX[ai][j]),P+1) + 
								pow(fabs(temp - RX[si][j]),P+1));

                SG[i][j] += WD*(pow(fabs(temp - RX[ai][aj]),P+1) + 
								 pow(fabs(temp - RX[ai][sj]),P+1) + 
								 pow(fabs(temp - RX[si][aj]),P+1) + 
								 pow(fabs(temp - RX[si][sj]),P+1));

                SG[i][j] /= 2.0;
				if(SG[i][j] < 1.0e-6)
					SG[i][j] = 1.0e-6;
        }
}

static double cal_SIGMA(float **X, int **mask, int flag_l_like)
{
    int i,j;
    int aj,sj,ai;
    double sum = 0.0, sumd = 0.0;
	double temp;

    for(j=0;j<M;j++)
        for(i=0;i<N;i++)
			if(mask[i][j] || flag_l_like)
			{
                ai = (i+1)%N;
                aj = (j+1)%M;
                sj = j-1;
                if (sj < 0)
                    sj = M-1;
    
    
                temp = (SG[i][j] + SG[ai][j])/2.0;
                sum += pow(fabs(X[i][j] - X[ai][j]),P+1)/temp;
                temp = (SG[i][j] + SG[i][aj])/2.0;
                sum += pow(fabs(X[i][j] - X[i][aj]),P+1)/temp;
                temp = (SG[i][j] + SG[ai][sj])/2.0;
                sumd += pow(fabs(X[i][j] - X[ai][sj]),P+1)/temp;
                temp = (SG[i][j] + SG[ai][aj])/2.0;
                sumd += pow(fabs(X[i][j] - X[ai][aj]),P+1)/temp;
         	}

    sum = sum*WN;
    sumd = sumd*WD;

    return((sum+sumd));
}




static void relax(float **X, int **mask, float **Y, struct sino_info *sino, struct geom_info *geom, float **profiles, int no_iter)
{

	int iter;
	int i,j,ii,jj,NN,MM,itemp,pixel,pro,t;
	int ai,aj,si,sj;
	float diff;
	double l_like, temp;
	char c;
	int FLAG;





	if (SIGMA == 0.0)
	{
		SIGMA = cal_SIGMA(X,mask,0)/Nsup;
		if (SIGMA == 0.0)
			SIGMA = 1.0;
	} 

	if(VERBOSE && param)
      printf("Sigma initialized to %f\n",pow(SIGMA,1.0/(P+1)));

	for(iter=0;iter<no_iter;iter++)
    {
        if((iter/4)%2 != 0)
        {
          NN = M;
          MM = N;
        }
        else
        {
          NN = N;
          MM = M;
        }

        for (ii=0; ii<NN; ii++)
            for(jj=0; jj<MM; jj++)
            {

                if((iter-1)%4 == 0)
                {
                  i = NN-1-ii;
                  j = MM-1-jj;
                }
                else if((iter-2)%4 == 0)
                {
                  i = NN-1-ii;
                  j = jj;
                }
                else if((iter-3)%4 == 0)
                {
                  i = ii;
                  j = MM-1-jj;
                }
                else
                {
                  i = ii;
                  j = jj;
                }

                if((iter/4)%2 != 0)
                {
                  itemp = i;
                  i = j;
                  j = itemp;
                }


				A_comp(i,j,sino, geom, profiles,&sp_col);
                if(sp_col.n_row)
                {


                    D1 = 0.0;
                    D2 = 0.0;
					FLAG = 0;
                    for(t=0;t<sp_col.n_row;t++)
                    {
                        pro = sp_col.row[t];
                        if (S[pro] != 0.0 && Y[0][pro] != 0)
                        {
                            D1 += sp_col.val[t]*(1.0 - Y[0][pro]/S[pro]);
                            D2 += Y[0][pro]*sq(sp_col.val[t]/S[pro]);

                        }
                        else if(Y[0][pro] == 0.0)
                            D1 += sp_col.val[t];
                        else
                            FLAG = 1;



                    }
                    D1 *= sp_col.multiplier;
                    D2 *= sq(sp_col.multiplier);





                    CX = X[i][j];
                    ML = CX - D1/D2;

                    ai = (i+1)%N;
                    aj = (j+1)%M;
                    si = i-1;
                    if (si < 0)
                        si = N-1;
                    sj = j-1;
                    if (sj < 0)
                        sj = M-1;

					SXN[0] = (SG[i][j] + SG[i][aj])/2.0;
					SXN[1] = (SG[i][j] + SG[i][sj])/2.0;
					SXN[2] = (SG[i][j] + SG[ai][j])/2.0;
					SXN[3] = (SG[i][j] + SG[si][j])/2.0;

					SXD[0] = (SG[i][j] + SG[ai][aj])/2.0;
					SXD[1] = (SG[i][j] + SG[ai][sj])/2.0;
					SXD[2] = (SG[i][j] + SG[si][aj])/2.0;
					SXD[3] = (SG[i][j] + SG[si][sj])/2.0;

                    NX[0] = X[i][aj];
                    NX[1] = X[i][sj];
                    NX[2] = X[ai][j];
                    NX[3] = X[si][j];

                    DX[0] = X[ai][aj];
                    DX[1] = X[ai][sj];
                    DX[2] = X[si][aj];
                    DX[3] = X[si][sj];


					X[i][j] = updateX();
                    if ( FLAG && X[i][j] == 0.0)
                      X[i][j] = 1.0/sp_col.n_row;
					diff = X[i][j] - CX;
                	if(diff)
                    	for(t=0;t<sp_col.n_row;t++)
                    	{
                        	pro = sp_col.row[t];
                        	S[pro] += sp_col.multiplier*sp_col.val[t]*diff;
                    	}

				}



            }

        if (param)
        {
			SIGMA = cal_SIGMA(X,mask,0)/Nsup;
			if(VERBOSE)
              printf("%d. sigma %f\n",iter+1,pow(SIGMA,1.0/(P+1)));
        }
        else
        {
            l_like = 0.0;
            for(i=0;i<Mpro;i++)
                if ( S[i] > 0.0 )
                    l_like += -S[i] + Y[0][i]*log( S[i] );

            l_like -= cal_SIGMA(X,mask,1)/((P+1)*SIGMA);
			if(VERBOSE)
			  printf("%d. Normalized log-likelihood %f\n",iter+1,l_like/(N*M));
        }

	}
	if(param )
	  printf("Estimated value of sigma %f\n",pow(SIGMA,1.0/(P+1)));


}	




static void minmax(float *low, float *high)
{
    int i;

    *low = ML;
    *high = ML;

    for(i=0;i<4;i++)
    {

        if( *low > NX[i])
            *low = NX[i];
        else if(*high < NX[i])
            *high = NX[i];

        if( *low > DX[i])
            *low = DX[i];
        else if(*high < DX[i])
            *high = DX[i];

    }
}


static float updateX(void)
{
	float mm,new,std,var,beta,temp;
	float correction,dE,pert,threshold,low,high;
	int status;


    minmax(&low, &high);

    mm = solve(df, low, high, TOL, &status);
    if (status)
    {
        if (f(low) < f(high))
            mm = low;
        else
            mm = high;
    }

	if (param)
	{
		var = 1/D2;
		if (var < 0.0)
		{	
            fprintf(stderr,"Second derivative of data model is negative!!\n");
			exit(-1);
		}
		std = sqrt(var);
		if ( mm > -std/16.0 || (temp=df(0.0)) <= 0.0 )
		{
			while((new=std*normal() + mm)<0.0);
			correction = (CX-mm)*(CX-mm)/(2.0*var);
			correction -= (new-mm)*(new-mm)/(2.0*var);
		}
		else
		{
			beta = 1.0/temp;
			new = beta*exprand();
			correction = (CX-new)/beta;
		}

		dE = f(new) - f(CX) + correction;

		if (!(dE < 0.0 || (random2() < exp(-dE))))
			new = CX;

	}
	else
		if( mm < 0.0)
			new = 0.0;
		else
			new = mm;

	return(new);
}
	
	

static double df(double x)
{
    float prior=0.0, dprior=0.0;
    int i;

    for(i=0;i<4;i++)
    {
        prior += pow(fabs(x - NX[i]),P)*SIGN(x,NX[i])/SXN[i];
        dprior += pow(fabs(x - DX[i]),P)*SIGN(x,DX[i])/SXD[i];
    }
    prior  = prior*WN;
    dprior  = dprior*WD;

    return( D1 + D2*(x - CX) + (prior+dprior)/SIGMA );
}

static double f(double x)
{
    float temp, prior=0.0, dprior=0.0;
    int i;

    temp = x-CX;

    for(i=0;i<4;i++)
    {
        prior += pow(fabs(x - NX[i]),P+1)/SXN[i];
        dprior += pow(fabs(x - DX[i]),P+1)/SXD[i];
    }
    prior  = prior*WN;
    dprior  = dprior*WD;

    return( D1*temp + D2*temp*temp/2.0 + (prior + dprior )/((P+1)*SIGMA) );
}


static void f_pro(float **X, struct sino_info *sino, struct geom_info *geom, float **profiles)
{
    int i,j,t;

    /* Adding scatter */
    for(i=0;i<Mpro;i++)
        S[i] = scatter;

    for(i=0;i<N;i++)
    for(j=0;j<M;j++){
		A_comp(i,j,sino,geom,profiles,&sp_col);
        for(t=0;t<sp_col.n_row; t++)
            S[sp_col.row[t]] += sp_col.multiplier*sp_col.val[t]*X[i][j];
	}
}


#undef TOL
#undef WN
#undef WD
#undef SIGN

