/* smap using new model (adaptive scaling parameter) */
#include <stdio.h>
#include <math.h>
#include "tdf/tdf.h"
#include "allocate.h"
#include "assort.h"
#include "solve.h"
#include "randlib.h"
#include "mr_trans.h"

#define TOL 0.00001
#define WN 0.14644660940673
#define WD 0.10355339059327

#define SG(A,B) ((A) > (B) ? 1.0 : -1.0)



static void f_pro(float **X, struct system_matrix *A);
static double f(double x);
static double df(double x);
static float updateX(void);
static double cal_SX(float **X, float **RX, struct system_matrix *A);
static void minmax(float *low, float *high);
static void set_SG(float **RX,struct system_matrix *A);
static void relax(float **X, float **RX, float **Y, float **YT, float **bias, struct system_matrix *A, int no_iter);


static float *S,**SG;
static int param, Nsup;

static double SIGMA, ML,IX,CX,SX;
static double D1, D2;


float **mr_trans(float **X, float **RX, float **Y, float **YT, float **bias, struct system_matrix *A, int no_iter_param, int no_iter_rec, float *start_SX)
{

	int N,M,i,j;

	srandom2(1);

	SIGMA = *start_SX;

	N = A->n_y;
	M = A->n_x;

	/* Allocate memory and initialize projection matrices*/
    S = (float *) get_spc(A->Mpro, sizeof(float));
    SG = (float **) get_img(M,N, sizeof(float));

  	Nsup = 0;
   	for( i=0;i<A->Npix;i++ )
      	if ( A->n_row[i] )
           	Nsup++;
		else
		{
			X[0][i] = 0.0;
			RX[0][i] = 0.0;
		}

	f_pro(X,A);

	param = 1;
	set_SG(RX,A);
	relax(X,RX,Y,YT,bias,A,no_iter_param);

	for(i=0;i<A->Npix;i++)
		X[0][i] = RX[0][i];
	f_pro(X,A);

	param = 0;
	relax(X,RX,Y,YT,bias,A,no_iter_rec);



	free((void *)S);

	*start_SX = SIGMA;
	return(SG);
}

		
static void set_SG(float **RX, struct system_matrix *A)
{
    int i,j,pixel,ai,aj,si,sj,N,M;
    float temp;

    N = A->n_y;
    M = A->n_x;


    for (i=0; i<N; i++)
        for(j=0; j<M; j++)
        {
            pixel = i*M+j;
			SG[0][pixel] = 0.0;
            if(A->n_row[pixel])
            {

                ai = (i+1)%N;
                aj = (j+1)%M;
                si = i-1;
                if (si < 0)
                    si = N-1;
                sj = j-1;
                if (sj < 0)
                    sj = M-1;

                temp = RX[i][j];
                SG[0][pixel] = WN*(fabs(temp - RX[i][aj]) + fabs(temp - RX[i][sj]) + fabs(temp - RX[ai][j]) + fabs(temp - RX[si][j]));

                SG[0][pixel] += WD*(fabs(temp - RX[ai][aj]) + fabs(temp - RX[ai][sj]) + fabs(temp - RX[si][aj]) + fabs(temp - RX[si][sj]));

				if(SG[0][pixel] == 0.0)
				{
					SG[0][pixel] = 1.0e-6;
					printf("Warning - SG 0.0\n");
				}
            }
        }
}




static void relax(float **X, float **RX, float **Y, float **YT, float **bias, struct system_matrix *A, int no_iter)
{

	int iter,status;
	int i,N,M,pro,t;
	float diff,high,x;
	double p_hat,lamda,A_lamda, l_like, temp;


	N = A->n_y;
	M = A->n_x;

	if(SIGMA == 0.0)
	{
		SIGMA = 1.0;
		printf("Warning - SIGMA arbitrarily initialized to 1.0\n");
	}



	for(iter=0;iter<no_iter;iter++)
    {

        for (i=0; i<A->Npix; i++)
            {

                if(A->n_row[i])
                {


                    D1 = 0.0;
                    D2 = 0.0;
                    for(t=0;t<A->n_row[i];t++)
                    {
                        pro = A->row[i][t];
						if (S[pro] >= 0.0)
						{

                      		p_hat = exp(-S[pro]);
                      		lamda = YT[0][pro]*p_hat;
                      		A_lamda = lamda+bias[0][pro];
                       		temp = A->val[i][t]*lamda;
                       		D1 += temp*(Y[0][pro]/A_lamda - 1.0);
                       		D2 += A->val[i][t]*temp*(1.0-bias[0][pro]*Y[0][pro]/(A_lamda*A_lamda));
						}

                    }



                    CX = X[0][i];
                    ML = CX - D1/D2;
                    IX = RX[0][i];
					SX = SIGMA*SG[0][i];

					X[0][i] = updateX();
					diff = X[0][i] - CX;
                	if(diff)
                    	for(t=0;t<A->n_row[i];t++)
                    	{
                        	pro = A->row[i][t];
                        	S[pro] += A->val[i][t]*diff;
                    	}

                }
					



            }

        if (param)
        {
			SIGMA = cal_SX(X,RX,A)/Nsup;
            printf("SIGMA %f\n",SIGMA);
        }
        else
        {
            l_like = 0.0;
            for(i=0;i<A->Mpro;i++)
            {
                if (S[i] > 0.0)
                    temp = YT[0][i]*exp(-S[i]);
                else
                    temp = YT[0][i];
                l_like += -temp + Y[0][i]*log( temp+bias[0][i] );
            }
            l_like -= cal_SX(X,RX,A)/SIGMA;
            printf("l_like %f\n",l_like);
        }



	}	


							
}

static double cal_SX(float **X, float **RX, struct system_matrix *A)
{
    int i;
    double sum=0.0;


    for(i=0;i<A->Npix;i++)
        if(A->n_row[i])
            sum += fabs(X[0][i] - RX[0][i])/SG[0][i];

    return(sum);
}



static void minmax(float *low, float *high)
{
    int i;

    *low = ML;
    *high = ML;

    if( *low > IX)
        *low = IX;
    else if(*high < IX)
        *high = IX;

}


static float updateX(void)
{
	float mm,new,std,var,beta,temp;
	float correction,dE,pert,threshold,low,high;
	int status;


        minmax(&low, &high);

        mm = solve(df, low, high, TOL, &status);
        if (status)
        {
            if (f(low) < f(high))
                mm = low;
            else
                mm = high;
        }

	if (param)
	{
		var = 1/D2;
		if (var < 0.0)
		{	
			printf("var. negative\n");
			exit(-1);
		}
		std = sqrt(var);
		if ( mm > -std/4.0 )
		{
			while((new=std*normal() + mm)<0.0);
			correction = (CX-mm)*(CX-mm)/(2.0*var);
			correction -= (new-mm)*(new-mm)/(2.0*var);
		}
		else
		{
			if (df(0.0) <= 0.0)
			{	
				printf("df negative %f,%f,%f,%f\n",df(0.0),mm,ML,std);
				exit(-1);
			}

			beta = 10.0/(df(0.0));
			new = beta*exprand();
			correction = (CX-new)/beta;
		}

		dE = f(new) - f(CX) + correction;

		if (!(dE < 0.0 || (random2() < exp(-dE))))
			new = CX;

	}
	else
		if( mm < 0.0)
			new = 0.0;
		else
			new = mm;

	return(new);
}
	
	


static double df(double x)
{
    return( D1 + D2*(x - CX) + SG(x,IX)/SX);

}

static double f(double x)
{
	float temp;

    temp = x-CX;
    return( D1*temp + D2*temp*temp/2.0 + fabs(x-IX)/SX);
}



static void f_pro(float **X, struct system_matrix *A)
{
    int i,j;

    for(i=0;i<A->Mpro;i++)
        S[i] = 0.0;

    for(j=0;j<A->Npix;j++)
        for(i=0;i<A->n_row[j]; i++)
            S[A->row[j][i]] += A->val[j][i]*X[0][j];
}


#undef TOL
#undef WN
#undef WD
#undef SG

