#include <stdio.h>
#include <math.h>
#include "A_comp.h"
#include "allocate.h"
#include "assort.h"
#include "solve.h"
#include "randlib.h"

#define TOL 1.0e-9
#define WN 0.14644660940673
#define WD 0.10355339059327

#define PN 5

static double proj(double *x, double *y, int n);
static double Q(double x);


#define SG(A,B) ((A) > (B) ? 1.0 : -1.0)

static void f_pro(float **X, struct sino_info *sino, struct geom_info *geom, float **profiles);
static void relax(float **X, int **mask, float **Y, struct sino_info *sino, struct geom_info *geom, float **profiles, int no_iter);
static double f(double x);
static double df(double x);

static float updateX(void);
static void minmax(float *low, float *high);


static float *S;
static int param, Nsup, Mpro,N,M;

static float CX,SX=0.0,ML,NX[4],DX[4],P;
static double D1, D2;
extern struct column sp_col;

extern int VERBOSE;

extern double scatter;


float mrf_emis_otf(float **X, int **mask, float **Y, struct sino_info *sino, struct geom_info *geom, int no_iter_param, int no_iter_rec, float p)
{


	int k,i,j;
	float *Xinit;
	float **profiles;



	srandom2(1);
	P = p-1;

	pix_prof_comp(sino,geom,&profiles);

	N = geom->n_y;
	M = geom->n_x;
	Mpro = sino->n_t*sino->n_theta;

    Nsup=0;
    for(i=0;i<N*M;i++)
      if(mask[0][i])
        Nsup++;


	/* Allocate memory*/


    S = (float *) get_spc(Mpro, sizeof(float));

    if( no_iter_param != 0)
    {
      Xinit = (float *) get_spc(N*M, sizeof(float));
      for( i=0;i<N*M;i++ )
        Xinit[i] = X[0][i];
      f_pro(X,sino,geom,profiles);
      SX = energy8_sup(X, mask, N, M,P+1)/Nsup;
      if (SX == 0.0)
        SX = mean(X,N,M);
      if(VERBOSE)
      {
        printf("Sigma initialized to %f\n",pow(SX,1.0/p));
        printf("Starting estimation of sigma ...\n");
      }
      param = 1;
      relax(X, mask, Y, sino, geom, profiles, no_iter_param);

      if( no_iter_rec != 0)
        for( i=0;i<N*M;i++ )
          X[0][i] = Xinit[i];
    }
    else
    {
      SX = energy8_sup(X, mask, N, M,P+1)/Nsup;
      printf("You have set the no. of iterations for parameter estimation to be 0\n");
      printf("Sigma computed from the initial condition is %f\n",pow(SX,1.0/p));
      printf("Enter desired value for sigma: \n");
      scanf("%f",&SX);
	  SX = pow(SX,p);
    }


    if(no_iter_rec != 0)
    {
      f_pro(X,sino,geom,profiles);
      if(VERBOSE)
      {
        printf("Sigma is %f\n",pow(SX,1.0/p));
        printf("Starting reconstruction ...\n");
      }

      param = 0;
      relax(X, mask, Y, sino, geom, profiles, no_iter_rec);
	  printf("Finished reconstruction\n");
    }





    if( no_iter_param != 0)
	  free((void *)Xinit);
	free((void *)S);
	free_img((void **)profiles);

	return(pow(SX,1.0/(P+1)));

}

static void relax(float **X, int **mask, float **Y, struct sino_info *sino, struct geom_info *geom, float **profiles, int no_iter)
{

	int iter;
	int i,j,ii,jj,NN,MM,itemp,pixel,pro,t;
	int ai,aj,si,sj;
	float diff;
	double l_like, data_l_like, temp;
	char c;
    double x[PN],y[PN],proj_sigma;
    int count=0,pn;
	int FLAG;



	for(iter=0;iter<no_iter;iter++)
    {
        if((iter/4)%2 != 0)
        {
          NN = M;
          MM = N;
        }
        else
        {
          NN = N;
          MM = M;
        }

        for (ii=0; ii<NN; ii++)
            for(jj=0; jj<MM; jj++)
            {

                if((iter-1)%4 == 0)
                {
                  i = NN-1-ii;
                  j = MM-1-jj;
                }
                else if((iter-2)%4 == 0)
                {
                  i = NN-1-ii;
                  j = jj;
                }
                else if((iter-3)%4 == 0)
                {
                  i = ii;
                  j = MM-1-jj;
                }
                else
                {
                  i = ii;
                  j = jj;
                }

                if((iter/4)%2 != 0)
                {
                  itemp = i;
                  i = j;
                  j = itemp;
                }


				A_comp(i,j,sino,geom,profiles,&sp_col);

                if(sp_col.n_row)
                {


                    D1 = 0.0;
                    D2 = 0.0;
					FLAG = 0;
                    for(t=0;t<sp_col.n_row;t++)
                    {
                        pro = sp_col.row[t];
                        if (S[pro] != 0.0 && Y[0][pro] != 0)
                        {
                            D1 += sp_col.val[t]*(1.0 - Y[0][pro]/S[pro]);
                            D2 += Y[0][pro]*sq(sp_col.val[t]/S[pro]);

                        }
						else if( Y[0][pro] == 0)
                            D1 += sp_col.val[t];
						else
							FLAG = 1;

                    }
					D1 *= sp_col.multiplier;
					D2 *= sq(sp_col.multiplier);




                    CX = X[i][j];
                    ML = CX - D1/D2;

                    ai = (i+1)%N;
                    aj = (j+1)%M;
                    si = i-1;
                    if (si < 0)
                        si = N-1;
                    sj = j-1;
                    if (sj < 0)
                        sj = M-1;

                    NX[0] = X[i][aj];
                    NX[1] = X[i][sj];
                    NX[2] = X[ai][j];
                    NX[3] = X[si][j];

                    DX[0] = X[ai][aj];
                    DX[1] = X[ai][sj];
                    DX[2] = X[si][aj];
                    DX[3] = X[si][sj];


					X[i][j] = updateX();
					if ( FLAG && X[i][j] == 0.0)
					  X[i][j] = 1.0/sp_col.n_row;
					diff = X[i][j] - CX;
                	if(diff)
                    	for(t=0;t<sp_col.n_row;t++)
                    	{
                        	pro = sp_col.row[t];
                        	S[pro] += sp_col.multiplier*sp_col.val[t]*diff;
                    	}

				}



            }

        if (param)
        {
            x[count%PN] = (P+1)*SX;
            SX = energy8_sup(X, mask, N, M,P+1)/Nsup;
            y[count%PN] = (P+1)*SX - x[count%PN];
            count++;

            if(count > PN)
                pn = PN;
            else
                pn = count;
            if(pn > 1)
              proj_sigma = proj(x,y,pn)/(P+1);
            else
              proj_sigma = SX;

            if(VERBOSE)
              printf("%d. sigma %f, extrapolated sigma %f\n",iter+1,pow(SX,1.0/(P+1)),pow(proj_sigma,1.0/(P+1)));

        }
        else
        {
            l_like = 0.0;
            for(i=0;i<Mpro;i++)
                if ( S[i] > 0.0 )
                    l_like += -S[i] + Y[0][i]*log( S[i] );

            l_like -= energy8_toroid(X,N,M,P+1)/((P+1)*SX);
            if(VERBOSE)
              printf("%d. Normalized log-likelihood %f\n",iter+1,l_like/(N*M));
        }

	}

    if(param)
    {
      if(proj_sigma > 0.0 && no_iter >= 3)
        SX = proj_sigma;
      printf("Estimated value of sigma %f\n",pow(SX,1.0/(P+1)));
    }


}	




static void minmax(float *low, float *high)
{
    int i;

    *low = ML;
    *high = ML;

    for(i=0;i<4;i++)
    {

        if( *low > NX[i])
            *low = NX[i];
        else if(*high < NX[i])
            *high = NX[i];

        if( *low > DX[i])
            *low = DX[i];
        else if(*high < DX[i])
            *high = DX[i];

    }
}


static float updateX(void)
{
	float mm,new,std,var,beta,temp;
	float correction,dE,pert,threshold,low,high;
	int status;


    minmax(&low, &high);

    mm = solve(df, low, high, TOL, &status);
    if (status)
    {
        if (f(low) < f(high))
            mm = low;
        else
            mm = high;
    }

	if (param)
	{
		var = 1/D2;
		if (var < 0.0)
		{	
            fprintf(stderr,"Second derivative of data model is negative!!\n");
			exit(-1);
		}
		std = sqrt(var);
		if ( mm > -std/16.0 || (temp=df(0.0)) <= 0.0 )
		{
			while((new=std*normal() + mm)<0.0);
			correction = (CX-mm)*(CX-mm)/(2.0*var);
			correction -= (new-mm)*(new-mm)/(2.0*var);
		}
		else
		{
			beta = 1.0/temp;
			new = beta*exprand();
			correction = (CX-new)/beta;
		}

		dE = f(new) - f(CX) + correction;

		if (!(dE < 0.0 || (random2() < exp(-dE))))
			new = CX;

	}
	else
		if( mm < 0.0)
			new = 0.0;
		else
			new = mm;

	return(new);
}


static double df(double x)
{
    float prior=0.0, dprior=0.0;
    int i;

    for(i=0;i<4;i++)
    {
        prior += pow(fabs(x - NX[i]),P)*SG(x,NX[i]);
        dprior += pow(fabs(x - DX[i]),P)*SG(x,DX[i]);
    }
    prior  = prior*WN;
    dprior  = dprior*WD;

    return( D1 + D2*(x - CX) + (prior+dprior)/SX );
}

static double f(double x)
{
    float temp, prior=0.0, dprior=0.0;
    int i;

    temp = x-CX;

    for(i=0;i<4;i++)
    {
        prior += pow(fabs(x - NX[i]),P+1);
        dprior += pow(fabs(x - DX[i]),P+1);
    }
    prior  = prior*WN;
    dprior  = dprior*WD;

    return( D1*temp + D2*temp*temp/2.0 + (prior + dprior )/((P+1)*SX) );
}


static void f_pro(float **X, struct sino_info *sino, struct geom_info *geom, float **profiles)
{
    int i,j,t;

    /* Adding scatter */
    for(i=0;i<Mpro;i++)
        S[i] = scatter;

    for(i=0;i<N;i++)
    for(j=0;j<M;j++)
	{
		A_comp(i,j,sino,geom,profiles,&sp_col);
        for(t=0;t<sp_col.n_row; t++)
            S[sp_col.row[t]] += sp_col.multiplier*sp_col.val[t]*X[i][j];
	}
}

static double proj(double *x, double *y, int n)
{
    int i;
    double mx=0,my=0,mxy=0,mxsq=0,m,c;

    for(i=0;i<n;i++)
    {
        mx += x[i];
        my += y[i];
        mxy += x[i]*y[i];
        mxsq += x[i]*x[i];
    }
    mx /= n;
    my /= n;
    mxy /= n;
    mxsq /= n;

    m = (mxy - mx*my)/(mxsq - mx*mx);
    c = m*mx - my;

    return( c/m );
}




#undef TOL
#undef WN
#undef WD
#undef SG

