/* This program uses single resolution MRF model and 
computes the projection matrix entries on the fly (otf) */

#include <stdio.h>
#include <math.h>
#include "A_comp.h"
#include "allocate.h"
#include "assort.h"
#include "solve.h"
#include "randlib.h"

#define TOL 1.0e-9
#define WN 0.14644660940673
#define WD 0.10355339059327

#define PN 3
#define STD 0.015

static double proj(double *x, double *y, int n);
static double Q(double x);


#define SG(A,B) ((A) > (B) ? 1.0 : -1.0)

static void f_pro(float **X, struct sino_info *sino, struct geom_info *geom, float **profiles);
static void relax(float **X, int **mask, float **Y, struct sino_info *sino, struct geom_info *geom, float **profiles, int no_iter);
static double f(double x);
static double df(double x);

static float updateX(void);
static float updateX_met(void);
static void minmax(float *low, float *high);


static float *S;
static int param, Nsup, Mpro,N,M;
static float mm;

static float CX,SX=0.0 /*0.01132471919384 0.01164352469968 0543189634129*/,ML,NX[4],DX[4],P;
static double D1, D2;
extern struct column sp_col;


float mrf_emis_otf(float **X, int **mask, float **Y, struct sino_info *sino, struct geom_info *geom, int no_iter_param, int no_iter_rec, float p)
{


	int k,i,j;
	float *Xinit;
	float **profiles;



	srandom2(1);
	P = p-1;

	pix_prof_comp(sino,geom,&profiles);

	N = geom->n_y;
	M = geom->n_x;
	Mpro = sino->n_t*sino->n_theta;

    Nsup=0;
    for(i=0;i<N*M;i++)
      if(mask[0][i])
        Nsup++;


	/* Allocate memory*/


    S = (float *) get_spc(Mpro, sizeof(float));
    Xinit = (float *) get_spc(N*M, sizeof(float));

    if( no_iter_param != 0)
    {
      for( i=0;i<N*M;i++ )
        Xinit[i] = X[0][i];
      f_pro(X,sino,geom,profiles);
      param = 1;
      relax(X, mask, Y, sino, geom, profiles, no_iter_param);
	  printf("Finished parameter estimation\n");

      if( no_iter_rec != 0)
        for( i=0;i<N*M;i++ )
          X[0][i] = Xinit[i];
    }

    if(no_iter_rec != 0)
    {
      f_pro(X,sino,geom,profiles);
      param = 0;
      relax(X, mask, Y, sino, geom, profiles, no_iter_rec);
	  printf("Finished reconstruction\n");
    }





	free((void *)Xinit);
	free((void *)S);
	free_img((void **)profiles);

	return(SX);

}

static void relax(float **X, int **mask, float **Y, struct sino_info *sino, struct geom_info *geom, float **profiles, int no_iter)
{

	int iter;
	int i,j,ii,jj,NN,MM,itemp,pixel,pro,t;
	int ai,aj,si,sj;
	float diff;
	double l_like, data_l_like, temp;
	char c;
    double x[PN],y[PN],proj_sigma;
    int count=0,pn;
	int FLAG;
	FILE *fp;
	double aa;



    if(!param)
	{
            printf("Log_like for initial condition\n");
            l_like = 0.0;
            for(i=0;i<Mpro;i++)
                if ( S[i] > 0.0 )
                    l_like += -S[i] + Y[0][i]*log( S[i] );

/*
            l_like -= energy8_toroid(X,N,M,P+1)/((P+1)*SX);
            printf("Normalized l_like %f\n",l_like/(N*M));
            */

            printf("%f ",l_like/(N*M));
            data_l_like = -energy8_toroid(X,N,M,P+1)/((P+1)*SX);
            printf("%f ",data_l_like/(N*M));
            l_like += data_l_like;
            printf("%f\n",l_like/(N*M));

	}



	if (SX==0.0)
	{
		SX = energy8_sup(X, mask, N, M,P+1)/Nsup;
		if (SX == 0.0)
			SX = mean(X,N,M);
	} 


    printf("Starting sigma %f Nsup %d\n",pow(SX,1.0/(P+1)),Nsup);
	for(iter=0;iter<no_iter;iter++)
    {
        if((iter/4)%2 != 0)
        {
          NN = M;
          MM = N;
        }
        else
        {
          NN = N;
          MM = M;
        }

        for (ii=0; ii<NN; ii++)
            for(jj=0; jj<MM; jj++)
            {

                if((iter-1)%4 == 0)
                {
                  i = NN-1-ii;
                  j = MM-1-jj;
                }
                else if((iter-2)%4 == 0)
                {
                  i = NN-1-ii;
                  j = jj;
                }
                else if((iter-3)%4 == 0)
                {
                  i = ii;
                  j = MM-1-jj;
                }
                else
                {
                  i = ii;
                  j = jj;
                }

                if((iter/4)%2 != 0)
                {
                  itemp = i;
                  i = j;
                  j = itemp;
                }


				A_comp(i,j,sino,geom,profiles,&sp_col);

                if(sp_col.n_row)
                {


                    D1 = 0.0;
                    D2 = 0.0;
					FLAG = 0;
                    for(t=0;t<sp_col.n_row;t++)
                    {
                        pro = sp_col.row[t];
                        if (S[pro] != 0.0 && Y[0][pro] != 0)
                        {
                            D1 += sp_col.val[t]*(1.0 - Y[0][pro]/S[pro]);
                            D2 += Y[0][pro]*sq(sp_col.val[t]/S[pro]);

                        }
						else if( Y[0][pro] == 0)
                            D1 += sp_col.val[t];
						else
							FLAG = 1;
						/*
                        else if(Y[0][pro] != 0.0)
                            printf("Projection values don't agree !!\n");
							*/

                    }
					D1 *= sp_col.multiplier;
					D2 *= sq(sp_col.multiplier);




                    CX = X[i][j];
                    ML = CX - D1/D2;

                    ai = (i+1)%N;
                    aj = (j+1)%M;
                    si = i-1;
                    if (si < 0)
                        si = N-1;
                    sj = j-1;
                    if (sj < 0)
                        sj = M-1;

                    NX[0] = X[i][aj];
                    NX[1] = X[i][sj];
                    NX[2] = X[ai][j];
                    NX[3] = X[si][j];

                    DX[0] = X[ai][aj];
                    DX[1] = X[ai][sj];
                    DX[2] = X[si][aj];
                    DX[3] = X[si][sj];


					X[i][j] = updateX();

					if(i==64 && j==64 && iter == 0)
					{
					  printf("opening file \n");
					  fp = fopen("density","w");
					  fprintf(fp,"mm = %f\n",mm);
					  fprintf(fp,"var = %f\n",1/D2);
					  for(aa=0.0;aa<7.0;aa=aa+0.01)
						fprintf(fp,"%f %f\n",aa,f(aa));
					  printf("Done writing file \n");
					  fclose(fp);
					}

					if ( FLAG && X[i][j] == 0.0)
					  X[i][j] = 1.0/sp_col.n_row;
					diff = X[i][j] - CX;
                	if(diff)
                    	for(t=0;t<sp_col.n_row;t++)
                    	{
                        	pro = sp_col.row[t];
                        	S[pro] += sp_col.multiplier*sp_col.val[t]*diff;
                    	}

				}



            }

        if (param)
        {
            x[count%PN] = (P+1)*SX;
            SX = energy8_sup(X, mask, N, M,P+1)/Nsup;
            y[count%PN] = (P+1)*SX - x[count%PN];
            count++;

            if(count > PN)
                pn = PN;
            else
                pn = count;
            if(pn > 1)
              proj_sigma = proj(x,y,pn)/(P+1);
            else
              proj_sigma = SX;
            printf("sigma %f proj_sigma %f Nsup %d\n",pow(SX,1.0/(P+1)),pow(proj_sigma,1.0/(P+1)),Nsup);

        }
        else
        {
            l_like = 0.0;
            for(i=0;i<Mpro;i++)
                if ( S[i] > 0.0 )
                    l_like += -S[i] + Y[0][i]*log( S[i] );

/*
            l_like -= energy8_toroid(X,N,M,P+1)/((P+1)*SX);
            printf("Normalized l_like %f\n",l_like/(N*M));
			*/

            printf("%f ",l_like/(N*M));
            data_l_like = -energy8_toroid(X,N,M,P+1)/((P+1)*SX);
            printf("%f ",data_l_like/(N*M));
			l_like += data_l_like;
            printf("%f\n",l_like/(N*M));
        }

	}

}	


							
			




static void minmax(float *low, float *high)
{
    int i;

    *low = ML;
    *high = ML;

    for(i=0;i<4;i++)
    {

        if( *low > NX[i])
            *low = NX[i];
        else if(*high < NX[i])
            *high = NX[i];

        if( *low > DX[i])
            *low = DX[i];
        else if(*high < DX[i])
            *high = DX[i];

    }
}


static float updateX(void)
{
	float new,std,var,beta,temp;
	float correction,dE,pert,threshold,low,high;
	int status;


    minmax(&low, &high);

    mm = solve(df, low, high, TOL, &status);
    if (status)
    {
        if (f(low) < f(high))
            mm = low;
        else
            mm = high;
    }

	if (param)
	{
		var = 1/D2;
		if (var < 0.0)
		{	
			printf("var. negative\n");
			exit(-1);
		}
		std = sqrt(var);
		if ( mm > -std/4.0 )
		{
			while((new=std*normal() + mm)<0.0);
			correction = (CX-mm)*(CX-mm)/(2.0*var);
			correction -= (new-mm)*(new-mm)/(2.0*var);
		}
		else
		{
			if (df(0.0) <= 0.0)
			{	
				printf("df negative %f,%f,%f,%f\n",df(0.0),mm,ML,std);
				exit(-1);
			}

			beta = 10.0/(df(0.0));
			new = beta*exprand();
			correction = (CX-new)/beta;
		}

		dE = f(new) - f(CX) + correction;

		if (!(dE < 0.0 || (random2() < exp(-dE))))
			new = CX;

	}
	else
		if( mm < 0.0)
			new = 0.0;
		else
			new = mm;

	return(new);
}

static float updateX_met(void)
{
    float new;
    float correction,dE;




    while((new=STD*normal() + CX)<0.0);
    correction = -log(1.0 - Q(-CX/STD));
    correction += log(1.0 - Q(-new/STD));



    dE = f(new) - f(CX) + correction;

    if (!(dE < 0.0 || (random2() < exp(-dE))))
        new = CX;


    return(new);
}

	
	

static double df(double x)
{
    float prior=0.0, dprior=0.0;
    int i;

    for(i=0;i<4;i++)
    {
        prior += pow(fabs(x - NX[i]),P)*SG(x,NX[i]);
        dprior += pow(fabs(x - DX[i]),P)*SG(x,DX[i]);
    }
    prior  = prior*WN;
    dprior  = dprior*WD;

    return( D1 + D2*(x - CX) + (prior+dprior)/SX );
}

static double f(double x)
{
    float temp, prior=0.0, dprior=0.0;
    int i;

    temp = x-CX;

    for(i=0;i<4;i++)
    {
        prior += pow(fabs(x - NX[i]),P+1);
        dprior += pow(fabs(x - DX[i]),P+1);
    }
    prior  = prior*WN;
    dprior  = dprior*WD;

    return( D1*temp + D2*temp*temp/2.0 + (prior + dprior )/((P+1)*SX) );
}


static void f_pro(float **X, struct sino_info *sino, struct geom_info *geom, float **profiles)
{
    int i,j,t;

    /* Adding scatter */
    for(i=0;i<Mpro;i++)
        S[i] = 0.001;

    for(i=0;i<N;i++)
    for(j=0;j<M;j++)
	{
		A_comp(i,j,sino,geom,profiles,&sp_col);
        for(t=0;t<sp_col.n_row; t++)
            S[sp_col.row[t]] += sp_col.multiplier*sp_col.val[t]*X[i][j];
	}
}

static double proj(double *x, double *y, int n)
{
    int i;
    double mx=0,my=0,mxy=0,mxsq=0,m,c;

    for(i=0;i<n;i++)
    {
        mx += x[i];
        my += y[i];
        mxy += x[i]*y[i];
        mxsq += x[i]*x[i];
    }
    mx /= n;
    my /= n;
    mxy /= n;
    mxsq /= n;

    m = (mxy - mx*my)/(mxsq - mx*mx);
    c = m*mx - my;

    return( c/m );
}

static double Q(double x)
{
    static int first=1;
    static double erf[1500];
    FILE *fin;
    int index;

    x = fabs(x);

    if(first)
    {
        first = 0;

        fin = fopen ("/home/duotone/d/saquib/tomo_single/Q", "r");
        if (fin == NULL)
        {
            fprintf(stderr,"Cannot read file named %s\n","Q");
            exit (-1);
        }

        fread(erf, sizeof(double), 1500, fin);

        fclose(fin);
    }

    index = x/0.01;
    if (index >= 1500)
    {
    /*  printf("Q function cannot be calculated\n");*/
        return(0.0);
    }

    return( erf[index] + (erf[index+1]-erf[index])*(x-index*0.01)/0.01 );
}




#undef TOL
#undef WN
#undef WD
#undef SG

